//********************************************************************
//NP_Tasks : a non-preemptive tasking routine for Atari 8 bit in C
//Written for the CC65 C compiler.
//
//**Beta Version**
//Send complaints/ridicule and bugs to : danwinslow@cox.net
//or PM to danwinslow on atariage.
//**Beta Version**
//
//Implements a simple tasking scheme. Tasks must yield the processer
//voluntarly by periodically calling the np_yield() function. Tasks
//which do not do so ( including the main() task ) will block the 
//other tasks.
//
// Author : Dan Winslow, 2009
//          
//********************************************************************

#include <string.h>
#include "np_tasks.h"

#define RDY_TCB_OFFSET \
asm("pha"); asm("lda _task_index"); asm("clc"); \
asm("rol"); asm("rol"); asm("rol"); \
asm("rol"); asm("clc"); asm("tay"); asm("pla"); 

void np_task_wrapper( void );

static unsigned char task_index;  
static unsigned char next_index;

static unsigned char shutdown;

static unsigned char tasks[16*max_tasks];
                          //0-1   address of task function
                          //2     a
                          //3     x
                          //4     y
                          //5     flags
                          //6-7   parm stack pointer
                          //8     call stack pointer
                          //9     task status, 0 : Inactive, 1 : active
                          //10-15 free 

static unsigned long wake_clocks[max_tasks];
                          
static unsigned char *call_stack_offset[max_tasks];                          
static unsigned char call_stack        [256*max_tasks];                            
static unsigned char *parm_stack       [256*max_tasks];      

static unsigned int  from;  //these two get used in a variety of ways                   
static unsigned int  to;                     
static unsigned char tmp;

//*******************************************************************
//clock time in jiffies for wait routines
//
//*******************************************************************
void np_jiffies( unsigned long *j )
{
  *j=( ( (*(unsigned char *)(18))*65536) +
       ( (*(unsigned char *)(19)*256)  ) +
       ( (*(unsigned char *)(18))      )
	 );
}


//*******************************************************************
//set up tasking data
//must be called before any other tasking routines
//*******************************************************************
void np_init_tasking( void )
{
  memset(call_stack,0x0,sizeof(call_stack));
  memset(parm_stack,0x0,sizeof(parm_stack));
  memset(tasks,0x0,sizeof(tasks));
  for(from=0;from<max_tasks;from++)
  {
    call_stack_offset[from]=&(call_stack[from*256]);
  }
  task_index=0;
  next_index=task_index;
  tasks[(task_index*16)+9]=1;          //main task is always runnable
}

//*******************************************************************
//Sets up for a task. This never gets called for task 0 ( main thread )
//The data parameter is pushed as an address and is available to
//the task as its only parameter.
//*******************************************************************
unsigned char np_init_task( task_type task_ptr, void *data )
{

  unsigned char index=NP_TASK_TABLE_FULL;
  //this starts from 1 because task 0 is always main task
  for(from=1;from<max_tasks;from++)
  {
    if ( tasks[(from*16)+9] == 0 ) //is it a free slot?
    {
      index=from;                  //yes
      break;
    }
  }
  if ( index == NP_TASK_TABLE_FULL )  // we must be out of slots
  {
    return index;
  }

  //store value of actual task address for JSR built in the wrapper function
  //this will be JSR'd to by np_task_wrapper()
  *(unsigned int *)(&tasks[(index*16)])=(unsigned int)task_ptr;

  //set up parameter stack
  from=(unsigned int)&parm_stack[256*(index+1)];      // parm stack base
  from-=2;                                            // decrement 1 position to account for data parameter
  *((unsigned int *)from)=(unsigned int)data;         // transfer parm addr into parmstack base
                                                      // and store parm stack in TCB
  *(unsigned int *)(&tasks[(index*16)+6])=(unsigned int)from;

  //set up call stack pointer
  //this embeds a call to np_task_wrapper so first rts from yield will go there
  tasks[(index*16)+8]=0xFD;
  call_stack[(256*index)+255]=(((unsigned int)np_task_wrapper-1) & 0xFF00) >> 8;
  call_stack[(256*index)+254]=(((unsigned int)np_task_wrapper-1) & 0x00FF);

  tasks[(index*16)+9]=1;          //mark as runnable
  return index;
}

//*******************************************************************
//select next task to swap to. Task 0 is main thread and should always 
//be available. If no other task is runnable then task 0 is returned to
//*******************************************************************
void next_task( void )
{
  next_index=task_index;
  tmp=4;
  while(tmp)
  {
    next_index++;
    if ( next_index == max_tasks ) next_index = 0;
    if ( tasks[(next_index*16)+9]==1 )
    {
      return;  //yes, lets go
    }
    tmp--;
  }
}

static unsigned int ptr1save;
static unsigned int ptr2save;

//*******************************************************************
//Context switch. Swaps the task by copying current stack and regs to
//save area, and reloading from other task save area, pushing a fake 
//JSR onto stack and 'returning' via RTS.
//This ( two page copies ) is the main performance hit for this scheme.
//*******************************************************************
void np_yield( void )
{

  //do we have another task to switch to?
  next_task();

  //no, same task so just return
  if ( task_index == next_index ) return;

  //block interrupts
  asm("sei");

  //push regs          
  asm("pha");          
  asm("txa");          
  asm("pha");          
  asm("tya");          
  asm("pha");          // stack is y x a ...
  
  //rdy TCB offset      
  RDY_TCB_OFFSET;
  
  //save regs in control table
  asm("pla");                         
  asm("sta _tasks+4,y");
  asm("pla");                         
  asm("sta _tasks+3,y");
  asm("pla");                         
  asm("sta _tasks+2,y"); // stack is clean
  
  //save sp
  asm("lda sp");
  asm("sta _tasks+6,y");
  asm("lda sp+1");
  asm("sta _tasks+7,y");
  
  //save call stack pointer
  asm("tsx");                          
  asm("txa");
  asm("sta _tasks+8,y");

  //copy up active portion of stack
  to  = call_stack_offset[task_index];
  from= 0x100;
  asm("         lda _from");
  asm("         sta ptr1");
  asm("         lda _from+1");
  asm("         sta ptr1+1");
  asm("         lda _to");
  asm("         sta ptr2");
  asm("         lda _to+1");
  asm("         sta ptr2+1");
  asm("         tsx");         //move stack ptr to Y 
  asm("         txa");
  asm("         tay");         //so only copy active portion of stack
  asm("@MU1:    lda (ptr1),y");
  asm("         sta (ptr2),y");
  asm("         iny");
  asm("         bne @MU1");

  //currently we aren't worrying about ZP variables

  //swap to new task
  task_index=next_index;

  from  =call_stack_offset[task_index];
  to    =0x100;

  //rdy TCB offset      
  RDY_TCB_OFFSET;
  
  //load new call stack pointer
  asm("lda _tasks+8,y");
  asm("tax");
  asm("txs");
  
  //load new sp
  asm("lda _tasks+6,y");
  asm("sta sp");
  asm("lda _tasks+7,y");
  asm("sta sp+1");
  
  //load regs
  asm("lda _tasks+2,y");
  asm("pha");               // stack is a ...
  asm("lda _tasks+3,y");
  asm("tax");
  asm("lda _tasks+4,y");
  asm("tay");
  asm("pla");               // stack is clean

  //copy down active portion of stack
  asm("         lda _from");
  asm("         sta ptr1");
  asm("         lda _from+1");
  asm("         sta ptr1+1");
  asm("         lda _to");
  asm("         sta ptr2");
  asm("         lda _to+1");
  asm("         sta ptr2+1");
  asm("         tsx");
  asm("         txa");
  asm("         tay");
  asm("@MD1:    lda (ptr1),y");
  asm("         sta (ptr2),y");
  asm("         iny");
  asm("         bne @MD1");

  //allow interrupts
  asm("cli");

  //go!
  asm("rts");
  
}

//*******************************************************************
//This function serves as a wrapper to call the actual task proc, and
//sets up for a controlled return from the task proc to do cleanup
//*******************************************************************
#pragma optimize(push,off)
void np_task_wrapper( void )
{
  //load start of actual task into the JSR at
  //TASK_GO below and execute it

  //rdy TCB offset      
  RDY_TCB_OFFSET;

  //build & execute JSR to task
  asm("         lda     _tasks+0,y");
  asm("         sta     TASK_GO+1");  
  asm("         lda     _tasks+1,y");
  asm("         sta     TASK_GO+2");  
  asm("         jmp     TASK_GO"); //needed this for some reason or the compiler throws TASK_GO away
  asm("TASK_GO: jsr     $0000");   //jsr target gets modified to actual task address 
  //if we get here, the task has terminated ( returned from its top level function )
  //mark as non-runnable
  tasks[(task_index*16)+9]=0;

  //and yield, never to return
  np_yield();

  //should never get here
}
#pragma optimize(pop)

//*******************************************************************
//returns number of runnable tasks
//*******************************************************************
unsigned char np_task_count( void )
{
  unsigned char count=0;
  for(from=0;from<max_tasks;from++)
  {
    if ( tasks[(from*16)+9]==1 )
    {
      count++;
    }
  }
  return count;
}

//*******************************************************************
//sleeps and yields
//*******************************************************************
void np_sleep( int jiffies )
{
  unsigned long now;
  np_jiffies(&wake_clocks[task_index]);
  wake_clocks[task_index]+=jiffies;

  while( wake_clocks[task_index] > now )
  {
    np_yield();
    if ( shutdown ) break;
    np_jiffies(&now);
  }
}

//*******************************************************************
//global shutdown
//*******************************************************************
void np_shutdown( void )
{
  if ( task_index == 0)
  {
    shutdown=1;
    //need to go through all tasks and rig next JSR to be back to
    //np_task_wrapper
    //then wait for them to terminate
  }
}

//*******************************************************************
//informs a task it should terminate
//*******************************************************************
unsigned char np_terminate( void )
{
  return shutdown;
}


